@@ -42,12 +42,14 @@
 #else
 #include <sys/file.h>
 #endif
+#include <openat.h>
 
 #include "../gnulib/lib/xalloc.h"
 #include "../gnulib/lib/human.h"
 #include "../gnulib/lib/canonicalize.h"
-#include "closeout.h"
 #include <modetype.h>
+
+#include "closeout.h"
 #include "savedirinfo.h"
 #include "buildcmd.h"
 #include "dirname.h"
@@ -117,6 +119,12 @@ enum WdSanityCheckFatality
     NON_FATAL_IF_SANITY_CHECK_FAILS
   };
 
+
+int get_current_dirfd(void)
+{
+  return AT_FDCWD;
+}
+
 
 int
 main (int argc, char **argv)
@@ -148,9 +156,14 @@ main (int argc, char **argv)
     options.xstat = debug_stat;
 
 #ifdef DEBUG
-  fprintf (stderr, "cur_day_start = %s", ctime (&p->cur_day_start));
+  fprintf (stderr, "cur_day_start = %s", ctime (&options.cur_day_start));
 #endif /* DEBUG */
 
+  /* state.cwd_dir_fd has to be initialised before we call build_expression_tree()
+   * because command-line parsing may lead us to stat some files.
+   */
+  state.cwd_dir_fd = AT_FDCWD;
+  
   /* We are now processing the part of the "find" command line 
    * after the -H/-L options (if any).
    */
@@ -175,20 +188,27 @@ main (int argc, char **argv)
     }
   
 
-  starting_desc = open (".", O_RDONLY);
+  starting_desc = open (".", O_RDONLY
+#if defined O_LARGEFILE
+			|O_LARGEFILE
+#endif
+			);
   if (0 <= starting_desc && fchdir (starting_desc) != 0)
     {
       close (starting_desc);
       starting_desc = -1;
     }
+  assert(starting_desc >= 0);
+
   if (starting_desc < 0)
     {
       starting_dir = xgetcwd ();
       if (! starting_dir)
 	error (1, errno, _("cannot get current directory"));
     }
+  set_stat_placeholders(&starting_stat_buf);
   if ((*options.xstat) (".", &starting_stat_buf) != 0)
-    error (1, errno, _("cannot get current directory"));
+    error (1, errno, _("cannot stat current directory"));
 
   /* If no paths are given, default to ".".  */
   for (i = end_of_leading_options; i < argc && !looks_like_expression(argv[i], true); i++)
@@ -217,9 +237,10 @@ main (int argc, char **argv)
   return state.exit_status;
 }
 
-boolean is_fts_enabled()
+boolean is_fts_enabled(int *ftsoptions)
 {
   /* this version of find (i.e. this main()) does not use fts. */
+  *ftsoptions = 0;
   return false;
 }
 
@@ -442,6 +463,7 @@ wd_sanity_check(const char *thing_to_stat,
   
   *changed = false;
   
+  set_stat_placeholders(newinfo);
   if ((*options.xstat) (current_dir, newinfo) != 0)
     error (1, errno, "%s", thing_to_stat);
   
@@ -531,7 +553,8 @@ enum SafeChdirStatus
     SafeChdirFailStat,
     SafeChdirFailWouldBeUnableToReturn,
     SafeChdirFailChdirFailed,
-    SafeChdirFailNonexistent
+    SafeChdirFailNonexistent,
+    SafeChdirFailDestUnreadable
   };
 
 /* Safely perform a change in directory.  We do this by calling
@@ -558,7 +581,11 @@ safely_chdir_lstat(const char *dest,
   
   saved_errno = errno = 0;
 
-  dotfd = open(".", O_RDONLY);
+  dotfd = open(".", O_RDONLY
+#if defined O_LARGEFILE
+	       |O_LARGEFILE
+#endif
+	       );
 
   /* We jump back to here if wd_sanity_check()
    * recoverably triggers an alert.
@@ -569,6 +596,7 @@ safely_chdir_lstat(const char *dest,
   if (dotfd >= 0)
     {
       /* Stat the directory we're going to. */
+      set_stat_placeholders(statbuf_dest);
       if (0 == options.xstat(dest, statbuf_dest))
 	{
 	  statflag = true;
@@ -677,7 +705,7 @@ safely_chdir_lstat(const char *dest,
 		      goto fail;
 		    }
 		}
-	      
+
 	      close(dotfd);
 	      return SafeChdirOK;
 	    }
@@ -791,7 +819,11 @@ safely_chdir_nofollow(const char *dest,
     }
   
   errno = 0;
-  fd = open(dest, O_RDONLY|extraflags);
+  fd = open(dest, O_RDONLY
+#if defined O_LARGEFILE
+	    |O_LARGEFILE
+#endif
+	    |extraflags);
   if (fd < 0)
     {
       switch (errno)
@@ -801,7 +833,7 @@ safely_chdir_nofollow(const char *dest,
 	case ENOENT:
 	  return SafeChdirFailNonexistent;
 	default:
-	  return SafeChdirFailChdirFailed;
+	  return SafeChdirFailDestUnreadable;
 	}
     }
   
@@ -840,18 +872,23 @@ safely_chdir(const char *dest,
 	     enum ChdirSymlinkHandling symlink_follow_option,
 	     boolean *did_stat)
 {
+  enum SafeChdirStatus result;
+  
   /* We're about to leave a directory.  If there are any -execdir
    * argument lists which have been built but have not yet been
    * processed, do them now because they must be done in the same
    * directory.
    */
-  complete_pending_execdirs(get_eval_tree());
+  complete_pending_execdirs(get_current_dirfd());
 
-#if defined(O_NOFOLLOW)
-  if (options.open_nofollow_available)
-    return safely_chdir_nofollow(dest, direction, statbuf_dest, symlink_follow_option, did_stat);
+#if !defined(O_NOFOLLOW)
+  options.open_nofollow_available = false;
 #endif
-  return safely_chdir_lstat(dest, direction, statbuf_dest, symlink_follow_option, did_stat);
+  if (options.open_nofollow_available)
+    result = safely_chdir_nofollow(dest, direction, statbuf_dest, symlink_follow_option, did_stat);
+  else
+    result = safely_chdir_lstat(dest, direction, statbuf_dest, symlink_follow_option, did_stat);
+  return result;
 }
 
 
@@ -895,7 +932,9 @@ chdir_back (void)
 	fprintf(stderr, "chdir_back(): chdir(<starting-point>)\n");
 
       if (fchdir (starting_desc) != 0)
-	error (1, errno, "%s", starting_dir);
+	{
+	  error (1, errno, "%s", starting_dir);
+	}
     }
 }
 
@@ -984,7 +1023,7 @@ static void do_process_top_dir(char *pathname,
   (void) pstat;
   
   process_path (pathname, base, false, ".", mode);
-  complete_pending_execdirs(get_eval_tree());
+  complete_pending_execdirs(get_current_dirfd());
 }
 
 static void do_process_predicate(char *pathname,
@@ -994,7 +1033,7 @@ static void do_process_predicate(char *pathname,
 {
   (void) mode;
   
-  state.rel_pathname = base;
+  state.rel_pathname = base;	/* cwd_dir_fd was already set by safely_chdir */
   apply_predicate (pathname, pstat, get_eval_tree());
 }
 
@@ -1125,10 +1164,17 @@ process_path (char *pathname, char *name, boolean leaf, char *parent,
   /* Now we really need to stat the directory, even if we know the
    * type, because we need information like struct stat.st_rdev.
    */
-  if (get_statinfo(pathname, name, &stat_buf) != 0)
-    return 0;
+  if (0 == stat_buf.st_mode)
+    {
+      /* This call was made conditional on Sat Apr 14 16:01:01 2007,
+       * and at that time the test suite passed without it.  Omitting
+       * this stat call saves a lot of system calls.
+       */
+      if (get_statinfo(pathname, name, &stat_buf) != 0)
+	return 0;
 
-  state.have_stat = true;
+      state.have_stat = true;
+    }
   mode = state.type = stat_buf.st_mode;	/* use full info now that we have it. */
   state.stop_at_current_level =
     options.maxdepth >= 0
@@ -1219,6 +1265,7 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
   if (statp->st_nlink < 2)
     {
       subdirs_unreliable = true;
+      subdirs_left = 0;
     }
   else
     {
@@ -1258,7 +1305,7 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
        * yet been processed, do them now because they must be done in
        * the same directory.
        */
-      complete_pending_execdirs(get_eval_tree());
+      complete_pending_execdirs(get_current_dirfd());
       
       if (strcmp (name, "."))
 	{
@@ -1282,6 +1329,7 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
 		  /* If there is a link we need to follow it.  Hence 
 		   * the direct call to stat() not through (options.xstat)
 		   */
+		  set_stat_placeholders(&stat_buf);
 		  if (0 != stat(".", &stat_buf))
 		    break;	/* skip the assignment. */
 		}
@@ -1296,6 +1344,7 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
 	      break;
 	      
 	    case SafeChdirFailNonexistent:
+	    case SafeChdirFailDestUnreadable:
 	    case SafeChdirFailStat:
 	    case SafeChdirFailNotDir:
 	    case SafeChdirFailChdirFailed:
@@ -1354,6 +1403,7 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
 		  state.exit_status = 1; /* We know the result is wrong, now */
 		  options.no_leaf_check = true;	/* Don't make same
 						   mistake again */
+		  subdirs_unreliable = 1;
 		  subdirs_left = 1; /* band-aid for this iteration. */
 		}
 	      
@@ -1385,13 +1435,12 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
        * yet been processed, do them now because they must be done in
        * the same directory.
        */
-      complete_pending_execdirs(get_eval_tree()); 
+      complete_pending_execdirs(get_current_dirfd()); 
 
       if (strcmp (name, "."))
 	{
 	  enum SafeChdirStatus status;
 	  struct dir_id did;
-	  boolean did_stat = false;
 	  
 	  /* We could go back and do the next command-line arg
 	     instead, maybe using longjmp.  */
@@ -1406,6 +1455,7 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
 	      dir = parent;
 	    }
 	  
+	  did_stat = false;
 	  status = safely_chdir (dir, TraversingUp, &stat_buf, SymlinkHandleDefault, &did_stat);
 	  switch (status)
 	    {
@@ -1417,6 +1467,7 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
 	      return;
 	      
 	    case SafeChdirFailNonexistent:
+	    case SafeChdirFailDestUnreadable:
 	    case SafeChdirFailStat:
 	    case SafeChdirFailSymlink:
 	    case SafeChdirFailNotDir:
@@ -1441,4 +1492,12 @@ process_dir (char *pathname, char *name, int pathlen, struct stat *statp, char *
 	free (cur_path);
       free_dirinfo(dirinfo);
     }
+
+  if (subdirs_unreliable)
+    {
+      /* Make sure we hasn't used the variable subdirs_left if we knew
+       * we shouldn't do so.
+       */
+      assert(0 == subdirs_left || options.no_leaf_check);
+    }
 }
