@@ -28,8 +28,10 @@
 #include <sys/utsname.h>
 #endif
 
+#include <sys/time.h>
 #include "xalloc.h"
 #include "quotearg.h"
+#include "timespec.h"
 
 
 #if ENABLE_NLS
@@ -49,6 +51,7 @@
 #include <string.h>
 #include <limits.h>
 #include <assert.h>
+#include <openat.h>
 
 
 struct debug_option_assoc
@@ -63,7 +66,8 @@ static struct debug_option_assoc debugassoc[] =
     { "tree", DebugExpressionTree, "Display the expression tree" },
     { "search",DebugSearch, "Navigate the directory tree verbosely" },
     { "stat", DebugStat, "Trace calls to stat(2) and lstat(2)" },
-    { "opt",  DebugExpressionTree|DebugTreeOpt, "Show diagnostic information relating to optimisation" }
+    { "opt",  DebugExpressionTree|DebugTreeOpt, "Show diagnostic information relating to optimisation" },
+    { "exec", DebugExec,  "Show diagnostic information relating to -exec, -execdir, -ok and -okdir" }
   };
 #define N_DEBUGASSOC (sizeof(debugassoc)/sizeof(debugassoc[0]))
 
@@ -150,8 +154,6 @@ show_valid_debug_options(FILE *fp, int full)
 void
 usage (FILE *fp, int status, char *msg)
 {
-  size_t i;
-  
   if (msg)
     fprintf (fp, "%s: %s\n", program_name, msg);
   
@@ -161,6 +163,23 @@ usage (FILE *fp, int status, char *msg)
   if (0 != status)
     exit (status);
 }
+
+void 
+set_stat_placeholders(struct stat *p)
+{
+#if HAVE_STRUCT_STAT_ST_BIRTHTIME
+  p->st_birthtime = 0;
+#endif
+#if HAVE_STRUCT_STAT_ST_BIRTHTIMENSEC
+  p->st_birthtimensec = 0;
+#endif
+#if HAVE_STRUCT_STAT_ST_BIRTHTIMESPEC_TV_NSEC
+  p->st_birthtimespec.tv_nsec = -1;
+#endif
+#if HAVE_STRUCT_STAT_ST_BIRTHTIMESPEC_TV_SEC
+  p->st_birthtimespec.tv_sec = 0;
+#endif
+}
 
 
 /* Get the stat information for a file, if it is 
@@ -169,14 +188,22 @@ usage (FILE *fp, int status, char *msg)
 int
 get_statinfo (const char *pathname, const char *name, struct stat *p)
 {
-  if (!state.have_stat && (*options.xstat) (name, p) != 0)
+  /* Set markers in fields so we have a good idea if the implementation
+   * didn't bother to set them (e.g., NetBSD st_birthtimespec for MS-DOS 
+   * files)
+   */
+  if (!state.have_stat)
     {
-      if (!options.ignore_readdir_race || (errno != ENOENT) )
+      set_stat_placeholders(p);
+      if ((*options.xstat) (name, p) != 0)
 	{
-	  error (0, errno, "%s", pathname);
-	  state.exit_status = 1;
+	  if (!options.ignore_readdir_race || (errno != ENOENT) )
+	    {
+	      error (0, errno, "%s", pathname);
+	      state.exit_status = 1;
+	    }
+	  return -1;
 	}
-      return -1;
     }
   state.have_stat = true;
   state.have_type = true;
@@ -190,23 +217,23 @@ get_statinfo (const char *pathname, const char *name, struct stat *p)
  */
 int
 get_info (const char *pathname,
-	  const char *name,
 	  struct stat *p,
 	  struct predicate *pred_ptr)
 {
+  boolean todo = false;
+  
   /* If we need the full stat info, or we need the type info but don't 
    * already have it, stat the file now.
    */
-  (void) name;
   if (pred_ptr->need_stat)
-    {
-      return get_statinfo(pathname, state.rel_pathname, p);
-    }
-  if ((pred_ptr->need_type && (0 == state.have_type)))
-    {
-      return get_statinfo(pathname, state.rel_pathname, p);
-    }
-  return 0;
+    todo = true;
+  else if ((pred_ptr->need_type && (0 == state.have_type)))
+    todo = true;
+  
+  if (todo)
+    return get_statinfo(pathname, state.rel_pathname, p);
+  else
+    return 0;
 }
 
 /* Determine if we can use O_NOFOLLOW.
@@ -247,14 +274,15 @@ check_nofollow(void)
  * than ';' (singles only).  If there are any, run them (this will
  * have no effect if there are no arguments waiting).
  */
-void
-complete_pending_execdirs(struct predicate *p)
+static void
+do_complete_pending_execdirs(struct predicate *p, int dirfd)
 {
-#if defined(NEW_EXEC)
   if (NULL == p)
     return;
   
-  complete_pending_execdirs(p->pred_left);
+  assert(state.execdirs_outstanding);
+  
+  do_complete_pending_execdirs(p->pred_left, dirfd);
   
   if (p->pred_func == pred_execdir || p->pred_func == pred_okdir)
     {
@@ -274,13 +302,20 @@ complete_pending_execdirs(struct predicate *p)
 	}
     }
 
-  complete_pending_execdirs(p->pred_right);
-#else
-  /* nothing to do. */
-  return;
-#endif
+  do_complete_pending_execdirs(p->pred_right, dirfd);
 }
 
+void
+complete_pending_execdirs(int dirfd)
+{
+  if (state.execdirs_outstanding)
+    {
+      do_complete_pending_execdirs(get_eval_tree(), dirfd);
+      state.execdirs_outstanding = false;
+    }
+}
+
+
 
 /* Examine the predicate list for instances of -exec which have been
  * terminated with '+' (build argument list) rather than ';' (singles
@@ -290,7 +325,6 @@ complete_pending_execdirs(struct predicate *p)
 void
 complete_pending_execs(struct predicate *p)
 {
-#if defined(NEW_EXEC)
   if (NULL == p)
     return;
   
@@ -315,10 +349,6 @@ complete_pending_execs(struct predicate *p)
     }
 
   complete_pending_execs(p->pred_right);
-#else
-  /* nothing to do. */
-  return;
-#endif
 }
 
 
@@ -331,7 +361,7 @@ cleanup(void)
   if (eval_tree)
     {
       complete_pending_execs(eval_tree);
-      complete_pending_execdirs(eval_tree);
+      complete_pending_execdirs(get_current_dirfd());
     }
 }
 
@@ -349,7 +379,7 @@ fallback_stat(const char *name, struct stat *p, int prev_rv)
     case ENOTDIR:
       if (options.debug_options & DebugStat)
 	fprintf(stderr, "fallback_stat(): stat(%s) failed; falling back on lstat()\n", name);
-      return lstat(name, p);
+      return fstatat(state.cwd_dir_fd, name, p, AT_SYMLINK_NOFOLLOW);
 
     case EACCES:
     case EIO:
@@ -377,12 +407,14 @@ fallback_stat(const char *name, struct stat *p, int prev_rv)
 int 
 optionh_stat(const char *name, struct stat *p)
 {
+  set_stat_placeholders(p);
   if (0 == state.curdepth) 
     {
       /* This file is from the command line; deference the link (if it
        * is a link).  
        */
-      int rv = stat(name, p);
+      int rv;
+      rv = fstatat(state.cwd_dir_fd, name, p, 0);
       if (0 == rv)
 	return 0;		/* success */
       else
@@ -392,7 +424,7 @@ optionh_stat(const char *name, struct stat *p)
     {
       /* Not a file on the command line; do not dereference the link.
        */
-      return lstat(name, p);
+      return fstatat(state.cwd_dir_fd, name, p, AT_SYMLINK_NOFOLLOW);
     }
 }
 
@@ -403,7 +435,9 @@ optionh_stat(const char *name, struct stat *p)
 int 
 optionl_stat(const char *name, struct stat *p)
 {
-  int rv = stat(name, p);
+  int rv;
+  set_stat_placeholders(p);
+  rv = fstatat(state.cwd_dir_fd, name, p, 0);
   if (0 == rv)
     return 0;			/* normal case. */
   else
@@ -417,7 +451,8 @@ optionl_stat(const char *name, struct stat *p)
 int 
 optionp_stat(const char *name, struct stat *p)
 {
-  return lstat(name, p);
+  set_stat_placeholders(p);
+  return fstatat(state.cwd_dir_fd, name, p, AT_SYMLINK_NOFOLLOW);
 }
 
 
@@ -428,6 +463,7 @@ debug_stat (const char *file, struct stat *bufp)
 {
   ++stat_count;
   fprintf (stderr, "debug_stat (%s)\n", file);
+
   switch (options.symlink_handling)
     {
     case SYMLINK_ALWAYS_DEREF:
@@ -713,6 +749,26 @@ process_leading_options(int argc, char *argv[])
   return end_of_leading_options;
 }
 
+static struct timespec 
+now(void)
+{
+  struct timespec retval;
+  struct timeval tv;
+  time_t t;
+  
+  if (0 == gettimeofday(&tv, NULL))
+    {
+      retval.tv_sec  = tv.tv_sec;
+      retval.tv_nsec = tv.tv_usec * 1000; /* convert unit from microseconds to nanoseconds */
+      return retval;
+    }
+  t = time(NULL);
+  assert(t != (time_t)-1);
+  retval.tv_sec = t;
+  retval.tv_nsec = 0;
+  return retval;
+}
+
 void 
 set_option_defaults(struct options *p)
 {
@@ -744,8 +800,8 @@ set_option_defaults(struct options *p)
   
   p->do_dir_first = true;
   p->maxdepth = p->mindepth = -1;
-  p->start_time = time (NULL);
-  p->cur_day_start = p->start_time - DAYSECS;
+  p->start_time = now();
+  p->cur_day_start = p->start_time.tv_sec - DAYSECS;
   p->full_days = false;
   p->stay_on_filesystem = false;
   p->ignore_readdir_race = false;
@@ -773,3 +829,13 @@ set_option_defaults(struct options *p)
 
   set_follow_state(SYMLINK_NEVER_DEREF); /* The default is equivalent to -P. */
 }
+
+
+/* get_start_dirfd
+ *
+ * Returns the fd for the directory we started in.
+ */
+int get_start_dirfd(void)
+{
+  return starting_desc;
+}
