@@ -153,7 +153,6 @@ hack_gnulib_tool_output() {
         */mdate-sh | */texinfo.tex) dest=doc;;
         *) dest=build-aux;;
       esac
-      test -d "$dest" || mkdir "$dest"
       cp -fp "${gnulibdir}"/"$file" "$dest" || exit
     done
 
@@ -234,15 +233,11 @@ main() {
     ## Invoke gnulib-tool to import the code.
     local tool="${gnulibdir}"/gnulib-tool
 
-    if run_gnulib_tool "${tool}" && 
-	hack_gnulib_tool_output "${gnulibdir}" &&
-	refresh_output_files && 
-	update_version_file   ; then
-	echo Done.
-    else
-	echo FAILED >&2
-	exit 1
-    fi
+    run_gnulib_tool "${tool}" && 
+    hack_gnulib_tool_output "${gnulibdir}" &&
+    refresh_output_files && 
+    update_version_file  &&
+    echo Done.
 }
 
 main "$@"
