@@ -1,18 +1,20 @@
-/* buildcmd.h -- build command lines from a stream of arguments
-   Copyright (C) 2005, 2006 Free Software Foundation, Inc.
+/* buildcmd.[ch] -- build command lines from a stream of arguments
+   Copyright (C) 2005 Free Software Foundation, Inc.
 
-   This program is free software: you can redistribute it and/or modify
+   This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation, either version 3 of the License, or
-   (at your option) any later version.
-   
+   the Free Software Foundation; either version 2, or (at your option)
+   any later version.
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
+   USA.
 */
 /*
  * Written by James Youngman.
@@ -45,6 +47,9 @@ struct buildcmd_state
 
   /* to-do flag. */
   int todo;
+
+  /* Directory in which to perform the exec. */
+  int dirfd;
 };
 
 struct buildcmd_control
@@ -117,8 +122,8 @@ extern void bc_push_arg (const struct buildcmd_control *ctl,
 			 int initial_args);
 
 extern void  bc_init_state(const struct buildcmd_control *ctl,
-					 struct buildcmd_state *state,
-					 void *usercontext);
+			   struct buildcmd_state *state,
+			   void *usercontext);
 extern enum BC_INIT_STATUS bc_init_controlinfo(struct buildcmd_control *ctl);
 extern size_t bc_get_arg_max(void);
 extern void bc_use_sensible_arg_max(struct buildcmd_control *ctl);
