@@ -1,5 +1,5 @@
 /* xargs -- build and execute command lines from standard input
-   Copyright (C) 1990, 91, 92, 93, 94, 2000, 2003, 2005 Free Software Foundation, Inc.
+   Copyright (C) 1990, 91, 92, 93, 94, 2000, 2003, 2004, 2005, 2006, 2007 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -97,11 +97,6 @@
 
 #include "wait.h"
 
-/* States for read_line. */
-#define NORM 0
-#define SPACE 1
-#define QUOTE 2
-#define BACKSLASH 3
 
 #ifdef STDC_HEADERS
 #include <stdlib.h>
@@ -168,6 +163,7 @@ typedef int boolean;
 
 #include <xalloc.h>
 #include "closeout.h"
+#include "gnulib-version.h"
 
 void error PARAMS ((int status, int errnum, char *message,...));
 
@@ -189,6 +185,9 @@ static int lineno = 0;
 static struct buildcmd_state bc_state;
 static struct buildcmd_control bc_ctl;
 
+/* Did we already complain about NUL characters in the input? */
+static int nullwarning_given = 0;
+
 
 /* If nonzero, when this string is read on stdin it is treated as
    end of file.
@@ -297,6 +296,7 @@ get_char_oct_or_hex_escape(const char *s)
     }
   else
     {
+      p = NULL;			/* Silence compiler warning. */
       error(1, 0,
 	    _("Invalid escape sequence %s in input delimiter specification."),
 	    s);
@@ -407,7 +407,6 @@ main (int argc, char **argv)
   char *default_cmd = "/bin/echo";
   int (*read_args) PARAMS ((void)) = read_line;
   void (*act_on_init_result)(void) = noop;
-  int env_too_big = 0;
   enum BC_INIT_STATUS bcstatus;
   
   program_name = argv[0];
@@ -575,6 +574,7 @@ main (int argc, char **argv)
 
 	case 'v':
 	  printf (_("GNU xargs version %s\n"), version_string);
+	  printf (_("Built using GNU gnulib version %s\n"), gnulib_version);
 	  return 0;
 
 	default:
@@ -657,11 +657,19 @@ main (int argc, char **argv)
       if (isatty(STDIN_FILENO))
 	{
 	  fprintf(stderr,
-		  "\n"
-		  "Execution of xargs will continue now, and it will "
-		  "try to read its input and run commands; if this is "
-		  "not what you wanted to happen, please type the "
-		  "end-of-file keystroke.\n");
+		  _("\n"
+		    "Execution of xargs will continue now, and it will "
+		    "try to read its input and run commands; if this is "
+		    "not what you wanted to happen, please type the "
+		    "end-of-file keystroke.\n"));
+	  if (always_run_command)
+	    {
+	      fprintf(stderr,
+		      _("Warning: %s will be run at least once.  "
+			"If you do not want that to happen, then press "
+			"the interrupt keystroke.\n"),
+		      argv[optind]);
+	    }
 	}
     }
   
@@ -743,24 +751,34 @@ main (int argc, char **argv)
 static int
 read_line (void)
 {
+/* States for read_line. */
+  enum read_line_state 
+    {
+      NORM = 0,
+      SPACE = 1,
+      QUOTE = 2,
+      BACKSLASH = 3
+    };
   static boolean eof = false;
   /* Start out in mode SPACE to always strip leading spaces (even with -i).  */
-  int state = SPACE;		/* The type of character we last read.  */
+  enum read_line_state state = SPACE; /* The type of character we last read.  */
   int prevc;			/* The previous value of c.  */
   int quotc = 0;		/* The last quote character read.  */
   int c = EOF;
   boolean first = true;		/* true if reading first arg on line.  */
+  boolean seen_arg = false;      /* true if we have seen any arg (or part of one) yet */
   int len;
   char *p = linebuf;
   /* Including the NUL, the args must not grow past this point.  */
   char *endbuf = linebuf + bc_ctl.arg_max - bc_state.cmd_initial_argv_chars - 1;
-
+  
   if (eof)
     return -1;
   while (1)
     {
       prevc = c;
       c = getc (input_stream);
+      
       if (c == EOF)
 	{
 	  /* COMPAT: SYSV seems to ignore stuff on a line that
@@ -800,9 +818,16 @@ read_line (void)
 		lineno++;	/* For -l.  */
 	      if (p == linebuf)
 		{
-		  /* Blank line.  */
-		  state = SPACE;
-		  continue;
+		  if (seen_arg)
+		    {
+		      /* An empty argument, add it to the list as normal. */
+		    }
+		  else
+		    {
+		      /* Blank line.  */
+		      state = SPACE;
+		      continue;
+		    }
 		}
 	      *p++ = '\0';
 	      len = p - linebuf;
@@ -818,7 +843,12 @@ read_line (void)
 			     initial_args);
 	      return len;
 	    }
-	  if (!bc_ctl.replace_pat && ISSPACE (c))
+	  seen_arg = true;
+
+	  /* POSIX: In the POSIX locale, the separators are <SPC> and
+	   * <TAB>, but not <FF> or <VT>. 
+	   */
+	  if (!bc_ctl.replace_pat && ISBLANK (c))
 	    {
 	      *p++ = '\0';
 	      len = p - linebuf;
@@ -860,6 +890,7 @@ read_line (void)
 	  if (c == quotc)
 	    {
 	      state = NORM;
+	      seen_arg = true; /* Makes a difference for e.g. just '' or "" as the first arg on a line */
 	      continue;
 	    }
 	  break;
@@ -868,6 +899,17 @@ read_line (void)
 	  state = NORM;
 	  break;
 	}
+      
+      if ( (0 == c) && !nullwarning_given )
+	{
+	  /* This is just a warning message.  We only issue it once. */
+	  error (0, 0,
+		 _("warning: a NUL character occurred in the input.  "
+		   "It cannot be passed through in the argument list.  "
+		   "Did you mean to use the --null option?"));
+	  nullwarning_given = 1;
+	}
+
 #if 1
       if (p >= endbuf)
         {
@@ -1221,7 +1263,7 @@ Usage: %s [-0prtx] [--interactive] [--null] [-d|--delimiter=delim]\n\
        [-I replace-str] [-i[replace-str]] [--replace[=replace-str]]\n\
        [-n max-args] [--max-args=max-args]\n\
        [-s max-chars] [--max-chars=max-chars]\n\
-       [-P max-procs]  [--max-procs=max-procs] [[--show-limits]\n\
+       [-P max-procs]  [--max-procs=max-procs] [--show-limits]\n\
        [--verbose] [--exit] [--no-run-if-empty] [--arg-file=file]\n\
        [--version] [--help] [command [initial-arguments]]\n"),
 	   program_name);
